/*
    WOCFS - Write-Only enCrypted File System
    Copyright (C) 2011, Michael Rao

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
     any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "utils.hpp"

void generate_key(int size)
{
  gcry_sexp_t key_spec, key, pub_key, sec_key;
  int rc;
  
  char bks[1000];
  assert(size>=1024 && size<=8192);
  snprintf(bks,1000,"(genkey (rsa (nbits 4:%d)))",size);

  printf("generate RSA pair with %d bits...\n",size);

  rc=gcry_sexp_new(&key_spec,bks,0,1);
  if(rc) {printf("error in gcry_sexp_new(\"%s\")\n",bks);exit(1);}
  
  rc=gcry_pk_genkey(&key,key_spec);
  gcry_sexp_release( key_spec );
  if(rc) {printf("error in gcry_sexp_relase(\"%s\")\n",bks);exit(1);}
    
  pub_key=gcry_sexp_find_token(key,"public-key",0);
  assert(pub_key);

  sec_key=gcry_sexp_find_token(key,"private-key",0);
  assert(sec_key);

  char bf[100000];
  char bc[100000];
  char pass[10000];
  char pass2[10000];

  printf("pass: ");fflush(stdout);
  getpass(pass,10000);
  printf("retype pass: ");fflush(stdout);
  getpass(pass2,10000);

  if(strcmp(pass,pass2)) {
    printf("error pass!=pass2!\n");
    exit(1);
  }
  bzero(pass2,strlen(pass2));
 
  char hash[32];
  gcry_md_hash_buffer(GCRY_MD_SHA256,hash,pass,strlen(pass));
  bzero(pass,strlen(pass));

  int r;
  /*
    r=gcry_sexp_sprint(key,GCRYSEXP_FMT_ADVANCED,bf,100000);
    assert(r>0 && r<100000);
    printf("%s",bf);
  */

  r=gcry_sexp_sprint(pub_key,GCRYSEXP_FMT_ADVANCED,bf,100000);
  assert(r>0 && r<100000);
  FILE *out=fopen("wocfs.pub_key","w");
  fprintf(out,"%s",bf);
  fclose(out);
  printf("wocfs.pub_key writed\n");

  r=gcry_sexp_sprint(sec_key,GCRYSEXP_FMT_ADVANCED,bf,100000);
  assert(r>0 && r<100000);
  /*
    out=fopen(".wocfs.sec_key","w");
    fprintf(out,"%s",bf);
    fclose(out);
  */

  bzero(bf+strlen(bf),32);
  
  {
    gcry_error_t r;
    gcry_cipher_hd_t hd;
    r=gcry_cipher_open (&hd, GCRY_CIPHER_AES256, GCRY_CIPHER_MODE_CBC,GCRY_CIPHER_SECURE);
    assert(r==0);
    r=gcry_cipher_setkey(hd,hash,32);
    assert(r==0);
    r=gcry_cipher_encrypt(hd, bc,32*((strlen(bf)+31)/32),bf,32*((strlen(bf)+31)/32));
    assert(r==0);
    gcry_cipher_close(hd);
    out=fopen("wocfs.sec_key","w");
    fwrite(bc,1,32*((strlen(bf)+31)/32),out);
    fclose(out);
    printf("wocfs.sec_key writed\n");
  }

  bzero(bf,strlen(bf));
  bzero(bc,sizeof(bc));

  gcry_sexp_release(key);
  gcry_sexp_release(pub_key);
  gcry_sexp_release(sec_key);
}

int main(int ac, char **av)
{
  int size=2048;
  
  while(1) {
    if(ac<2) break;
    if(0==strcmp(av[1],"-n")) {size=atoi(av[2]);ac-=2;av+=2;continue;}
    break;
  }
  char *version;
  //gcry_control(GCRYCTL_DISABLE_SECMEM);
  version = strdup( gcry_check_version(NULL) );
  printf("Libgcrypt version: %s\n",version);

  /*
  char smash[100000];
  {
    int i=0;

    do {
      printf("Smash your keyborad and press enter... \n");
      echo_off();
      while(i<99999&&(smash[i]=getchar())!='\n') {
	i++;
	fprintf(stderr,"%d\r",i);
      }
      echo_on();
      smash[i]=0;
      printf("        \r");fflush(stdout);
    } while(i<2000);
  }
  */
  
  assert(size>=1024 && size<=8192);
  generate_key(size);

  printf( "OK: Exit.\n");
  return 0;
}
