/*
    WOCFS - Write-Only enCrypted File System
    Copyright (C) 2011, Michael Rao

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
     any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "opttree.hpp"
#include <map>

opttree_t *opttree_root=NULL;

struct wocfs_opts_t {
  const char* arg;
  int type;
  //void *ptr;
  long long value;
};

wocfs_opts_t opts_str[] ={
  {"unlink",POPT_MASK,A_UNLINK},
  {"plain-unlink",POPT_MASK,A_P_UNLINK},
  {"rename",POPT_MASK,A_RENAME},
  {"plain-rename",POPT_MASK,A_P_RENAME},
  {"create",POPT_MASK,A_CREATE},
  {"plain-create",POPT_MASK,A_P_CREATE},
  {"mkdir",POPT_MASK,A_MKDIR},
  {"rmdir",POPT_MASK,A_RMDIR},
  {"trunc",POPT_MASK,A_TRUNC},
  {"plain-trunc",POPT_MASK,A_P_TRUNC},
  {"openrw",POPT_MASK,A_ORDWR},
  {"symlink",POPT_MASK,A_P_SYMLNK},
  {"plain-symlink",POPT_MASK,A_P_SYMLNK},
  {"synconrelease",POPT_MASK,A_FINISHONRELEASE},
  {"fakeread",POPT_MASK,A_FAKEREAD},
  //{"compress",POPT_MASK,&global_opts,WO_ZLIB},
};

string o2str(long long o)
{
  string r;
  for(unsigned int i=0;i<sizeof(opts_str)/sizeof(opts_str[0]);i++) {
    if(o&opts_str[i].value)
      r=r+string(opts_str[i].arg);
  }
  return r;
}

void mopt_t::print() const 
{
  if(m1) 
    printf("+{%s} ",o2str(m1).c_str());
  if(m0) 
    printf("-{%s} ",o2str(m0).c_str());
}

bool vv2o(const char *arg, const char *v, mopt_t &o)
{
  if(0==strcmp(arg,"timeout")) {
    o.to=atoll(v);
    return 1;
  }
  if(0==strcmp(arg,"maxblock")) {
    o.mb=atoll(v);
  }

  fprintf(stderr,"Warn: Unkonwn option %s\n",arg);
  return 0;
}

bool f2o(const char *arg, mopt_t &o)
{
  for(unsigned int i=0;i<sizeof(opts_str)/sizeof(opts_str[0]);i++) {
    if(0==strcmp(arg,opts_str[i].arg)) {
      if(opts_str[i].type==POPT_MASK) 
	o.m1|=opts_str[i].value;
      return 1;
    }
    if(0==strncmp(arg,"no-",3) && 
       0==strcmp(arg+3,opts_str[i].arg)) {
      if(opts_str[i].type==POPT_MASK) 
	o.m0|=opts_str[i].value;
      return 1;
    }
  }

  fprintf(stderr,"Warn: Unkonwn option %s\n",arg);
  return 0;
}


extern int yyparse(void);
char *bufferfunction;
int posbuffct;

void affregex(opttree_t *t,int x=0) 
{
  while(t) {
    printf("%d | %s: ",x,t->test.expr.c_str());
    t->opt.print();
    printf("\n");
    affregex(t->yes,x+1);
    affregex(t->no,x+1);
    t=t->next;
  }
}

bool readopts(const char *fn)
{
  char buff[100000];
  FILE *in=fopen(fn,"r");
  if(in==NULL) {
    fprintf(stderr,"cannot open file %s\n",fn);
    return 0;
  }
  fread(buff,99999,1,in);
  bufferfunction=buff;
  posbuffct=0;
  yyparse();

  affregex(opttree_root);
  
  fclose(in);
  return 1;
}

map<string,opt_t> mopts,mopts_old; 

void clean_mopts() 
{
  mopts.swap(mopts_old);
  mopts.clear();
}

void compregex(opttree_t *t) 
{
  while(t) {
    if(t->test.expr!=string("")) {
      t->test.regex=new regex_t;
      regcomp(t->test.regex,t->test.expr.c_str(),REG_EXTENDED|REG_NOSUB);
    }
    compregex(t->yes);
    compregex(t->no);
    t=t->next;
  }
}

void getopt_(opttree_t *t,const char *path, opt_t &o) 
{
  while(t) {
    o.opt|=t->opt.m1;
    o.opt&=~t->opt.m0;
    if(t->opt.to>=-1) o.timeout=t->opt.to;
    if(t->opt.mb>=-1) o.maxbuffer=t->opt.mb;
    if(t->test.regex) {
      if(0==regexec(t->test.regex,path,0,NULL,0)) 
	getopt_(t->yes,path,o);
      else
	getopt_(t->no,path,o);
    } else {
      getopt_(t->yes,path,o);
    }
    
    t=t->next;
  }
}

#define max_mopts_size 200
    
ll_t getopt(const char *path) 
{
  map<string,opt_t>::iterator it=mopts.find(path);
  if(it!=mopts.end()) return it->second.opt;
  it=mopts_old.find(path);
  if(it!=mopts_old.end()) return it->second.opt;

  if(mopts.size()>max_mopts_size) clean_mopts();
    
  opt_t o=default_opts;

  getopt_(opttree_root,path,o);

  mopts[path]=o;
  
  return o.opt;
}


